/*****************************************************************************
 * objects.c: vlc_object_t handling
 *****************************************************************************
 * Copyright (C) 2002 VideoLAN
 * $Id: main.h,v 1.35 2002/04/27 22:11:22 gbazin Exp $
 *
 * Authors: Samuel Hocevar <sam@zoy.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 *****************************************************************************/

#include <vlc/vlc.h>

#ifdef HAVE_STDLIB_H
#   include <stdlib.h>                                          /* realloc() */
#endif

#include "stream_control.h"
#include "input_ext-intf.h"
#include "input_ext-dec.h"

#include "video.h"
#include "video_output.h"

#include "vdec_ext-plugins.h"

#include "audio_output.h"

#include "playlist.h"
#include "interface.h"

static vlc_object_t * vlc_object_find_inner( vlc_object_t *, int, int );
static void vlc_object_unlink_inner( vlc_object_t *, vlc_object_t * );


void vlc_dumpstructure_inner( int i_level, vlc_object_t *p_this )
{
    int i;
    char psz_foo[100];
    memset( psz_foo, ' ', 2 * i_level );
    psz_foo[2 * i_level] = '\0';

    fprintf(stderr, "%s* %i: %i children\n", psz_foo, p_this->i_type, p_this->i_children);

    for( i = 0 ; i < p_this->i_children ; i++ )
    {
        vlc_dumpstructure_inner( i_level + 1, p_this->pp_children[i] );
    }
}

void vlc_dumpstructure( vlc_object_t *p_this )
{
    vlc_dumpstructure_inner( 0, p_this->p_vlc->p_this );
}

/* vlc_object_init: initialize a vlc object and set its parent */
void * __vlc_object_create( vlc_object_t *p_this, int i_type )
{
    vlc_object_t *p_new;
    size_t i_size;

    switch( i_type )
    {
        case VLC_OBJECT_ROOT:
            i_size = sizeof(vlc_t);
            break;
        case VLC_OBJECT_MODULE:
            i_size = sizeof(module_t);
            break;
        case VLC_OBJECT_INTF:
            i_size = sizeof(intf_thread_t);
            break;
        case VLC_OBJECT_PLAYLIST:
            i_size = sizeof(playlist_t);
            break;
        case VLC_OBJECT_INPUT:
            i_size = sizeof(input_thread_t);
            break;
        case VLC_OBJECT_DECODER:
            i_size = sizeof(decoder_fifo_t);
            break;
        case VLC_OBJECT_VOUT:
            i_size = sizeof(vout_thread_t);
            break;
        case VLC_OBJECT_AOUT:
            i_size = sizeof(aout_thread_t);
            break;
        case VLC_OBJECT_VDEC:
            i_size = sizeof(vdec_thread_t);
            break;
        default:
            return NULL;
    }

    p_new = malloc( i_size );

    if( !p_new )
    {
        return NULL;
    }

    memset( p_new, 0, i_size );

    p_new->i_type = i_type;
    p_new->psz_type = "unknown";

    p_new->i_pid = 0;
    p_new->i_refcount = 0;
    p_new->b_die = 0;
    p_new->b_error = 0;
    p_new->b_active = 0;

    p_new->p_this = p_new;

    /* If i_type is root, then p_new is our own p_vlc */
    p_new->p_vlc = (i_type == VLC_OBJECT_ROOT) ? (vlc_t*)p_new : p_this->p_vlc;

    p_new->pp_parents = NULL;
    p_new->i_parents = 0;
    p_new->pp_children = NULL;
    p_new->i_children = 0;

    return p_new;
}

/* vlc_object_find: find a typed object and increment its refcount */
vlc_object_t * __vlc_object_find( vlc_object_t *p_this, int i_type, int i_mode )
{
    vlc_object_t *p_found;

    vlc_mutex_lock( &p_this->p_vlc->structure_lock );

    /* If we are of the requested type ourselves, don't look further */
    if( !(i_mode & FIND_STRICT) && p_this->i_type == i_type )
    {
        p_this->i_refcount++;
        vlc_mutex_unlock( &p_this->p_vlc->structure_lock );
        return p_this;
    }

    p_found = vlc_object_find_inner( p_this, i_type, i_mode );

    vlc_mutex_unlock( &p_this->p_vlc->structure_lock );

    return p_found;
}

static vlc_object_t * vlc_object_find_inner( vlc_object_t *p_this,
                                             int i_type, int i_mode )
{
    int i;
    vlc_object_t *p_tmp;

    switch( i_mode & 0x000f )
    {
    case FIND_PARENT:
        for( i = p_this->i_parents; i--; )
        {
            p_tmp = p_this->pp_parents[i];
            if( p_tmp->i_type == i_type )
            {
                p_tmp->i_refcount++;
                return p_tmp;
            }
            else if( p_tmp->i_parents )
            {
                p_tmp = vlc_object_find_inner( p_tmp, i_type, i_mode );
                if( p_tmp )
                {
                    return p_tmp;
                }
            }
        }
        break;

    case FIND_CHILD:
        for( i = p_this->i_children; i--; )
        {
            p_tmp = p_this->pp_children[i];
            if( p_tmp->i_type == i_type )
            {
                p_tmp->i_refcount++;
                return p_tmp;
            }
            else if( p_tmp->i_children )
            {
                p_tmp = vlc_object_find_inner( p_tmp, i_type, i_mode );
                if( p_tmp )
                {
                    return p_tmp;
                }
            }
        }
        break;

    case FIND_ANYWHERE:
        /* FIXME: unimplemented */
        break;
    }

    return NULL;
}

/* vlc_object_acquire: increment an object refcount */
void __vlc_object_acquire( vlc_object_t *p_this )
{
    vlc_mutex_lock( &p_this->p_vlc->structure_lock );
    p_this->i_refcount++;
    vlc_mutex_unlock( &p_this->p_vlc->structure_lock );
}

/* vlc_object_release: decrement an object refcount */
void __vlc_object_release( vlc_object_t *p_this )
{
    vlc_mutex_lock( &p_this->p_vlc->structure_lock );
    p_this->i_refcount--;
    vlc_mutex_unlock( &p_this->p_vlc->structure_lock );
}

/* vlc_object_unlink: detach object from its parents */
void __vlc_object_unlink_all( vlc_object_t *p_this )
{
    vlc_mutex_lock( &p_this->p_vlc->structure_lock );

    /* FIXME: BORK ! BORK ! BORK !!! THIS STUFF IS BORKED !! FIXME */
    while( p_this->i_parents )
    {
        /* Not very effective because we know the index, but we'd have to
         * parse p_parent->pp_children anyway. Plus, we remove duplicates
         * by not using the object's index */
        vlc_object_unlink_inner( p_this, p_this->pp_parents[0] );
    }

    vlc_mutex_unlock( &p_this->p_vlc->structure_lock );
}

/* vlc_object_unlink: remove a parent/child link */
void __vlc_object_unlink( vlc_object_t *p_this, vlc_object_t *p_parent )
{
    vlc_mutex_lock( &p_this->p_vlc->structure_lock );
    vlc_object_unlink_inner( p_this, p_parent );
    vlc_mutex_unlock( &p_this->p_vlc->structure_lock );
}

/* vlc_object_setparent: attach a parent object */
static void vlc_object_unlink_inner( vlc_object_t *p_this,
                                     vlc_object_t *p_parent )
{
    int i_index;

    for( i_index = p_this->i_parents ; i_index-- ; )
    {
        if( p_this->pp_parents[i_index] == p_parent )
        {
            memmove( p_this->pp_parents + i_index,
                     p_this->pp_parents + i_index + 1,
                     p_this->i_parents - i_index );
            p_this->i_parents--;
        }
    }

    for( i_index = p_parent->i_children ; i_index-- ; )
    {
        if( p_parent->pp_children[i_index] == p_this )
        {
            memmove( p_parent->pp_children + i_index,
                     p_parent->pp_children + (i_index + 1),
                     p_parent->i_children - (i_index + 1) );
            p_parent->i_children--;
        }
    }
}

void __vlc_object_setparent( vlc_object_t *p_this, vlc_object_t *p_parent )
{
    vlc_mutex_lock( &p_this->p_vlc->structure_lock );

    p_this->i_parents++;
    p_this->pp_parents = (vlc_object_t **)realloc( p_this->pp_parents,
                            p_this->i_parents * sizeof(vlc_object_t *) );
    p_this->pp_parents[p_this->i_parents - 1] = p_parent;

    p_parent->i_children++;
    p_parent->pp_children = (vlc_object_t **)realloc( p_parent->pp_children,
                               p_parent->i_children * sizeof(vlc_object_t *) );
    p_parent->pp_children[p_parent->i_children - 1] = p_this;

    vlc_mutex_unlock( &p_this->p_vlc->structure_lock );
}

/* vlc_object_setchild: attach a child object */
void __vlc_object_setchild( vlc_object_t *p_this, vlc_object_t *p_child )
{
    vlc_mutex_lock( &p_this->p_vlc->structure_lock );

    p_this->i_children++;
    p_this->pp_children = (vlc_object_t **)realloc( p_this->pp_children,
                             p_this->i_children * sizeof(vlc_object_t *) );
    p_this->pp_children[p_this->i_children - 1] = p_child;

    p_child->i_parents++;
    p_child->pp_parents = (vlc_object_t **)realloc( p_child->pp_parents,
                             p_child->i_parents * sizeof(vlc_object_t *) );
    p_child->pp_parents[p_child->i_parents - 1] = p_this;

    vlc_mutex_unlock( &p_this->p_vlc->structure_lock );
}

